// -*- c++ -*- for emacs
//
// Config -- Configuration parser and holder
//
// This object parses configuration files and creates a multi-layer hash
// that programs can query via methods.  It allows the setting and saving
// of values as well.  Configurations files are apache style
//
// <KeyType key>
//     key      information
// </KeyType>
//
// key   information
//
// example:
//
// Port  143
// <Domain urbanrage.com>
//       Database         /usr/db/database
//       <Server pop.urbanrage.com>
//              OS                    solaris-2.7
//              Architecture          sparc
//       </Server>
// </Domain>
//
//
// $Id: Config.cc,v 1.2 1999/02/27 23:18:41 eric Exp $
//

#include "Config.h"

// for perror
#include <stdio.h>
#include <errno.h>

#include <cstdarg>

#define __DEBUG (*__debug_ptr)
//static ofstream __debug_null("/dev/null");

void *
Config::GetItem(char* iName ...) {
  Conf_Map_t* l_map; 
  char*       l_str;
  char*       l_next_str;
  char*       l_last_str;
  va_list     l_ap;

  va_start(l_ap, iName);
  l_map = __current_map;
  for (l_str = iName; l_str != NULL; l_last_str = l_str, l_str = va_arg(l_ap, char*)) {
    switch (_isMap(l_map, l_str)) {
    case 0 : {
      l_next_str = va_arg(l_ap, char*);
      va_end(l_ap);
      if (l_next_str != 0) {
	return NULL;
      } else {
	return ((*l_map)[l_str]->GetValue());
      }
    } break;
    case 1 : {
      l_map = (*l_map)[l_str]->GetMap();
    } break;
    default : {
      va_end(l_ap);
      return NULL;
    } break;
    };
  }
  va_end(l_ap);
  return NULL; // only found a reference
}

bool
Config::SetItem(void* iItem, char* iName ... ) {
  Conf_Map_t* l_map;
  Conf_Map_t* l_last_map;
  Conf_Map_t* l_new_map;
  char*       l_str;
  char*       l_next_str;
  char*       l_last_str;
  va_list     l_ap;

  va_start(l_ap, iName);
  l_map = __current_map;
  for (l_str = iName; l_str != 0; l_last_str = l_str, l_str = va_arg(l_ap, char*)) {
    switch (_isMap(l_map, l_str)) {
    case 0 : {
      l_next_str = va_arg(l_ap, char*);
      va_end(l_ap);
      if (l_next_str != 0) { // trying to map through a value
	return false;
      } else {
	(*l_map)[l_str]->SetValue(iItem);
	return true;
      }
    } break;
    case -1 : {
      l_new_map = new Conf_Map_t;
      (*l_map)[l_str] = new ConfigItem(l_new_map);
    } break;
    default : {
      l_last_map = l_map;
      l_map = (*l_map)[l_str]->GetMap();
    }
    };
  }
  va_end(l_ap);
  if (_isMap(l_last_map, l_last_str)) { // it already exists as a map
    return false;
  }
  (*l_last_map)[l_last_str]->SetValue(iItem);
  return true;
}

bool
Config::SetLevel(char* iName ...) {
  Conf_Map_t* l_map;
  char*      l_str;
  char*      l_last_str;
  va_list    l_ap;

  va_start(l_ap, iName);
  l_map = &__root;
  for (l_str = iName; l_str != 0; l_last_str = l_str, l_str = va_arg(l_ap, char*)) {
    switch (_isMap(l_map, l_str)) {
    case 0 : {
      va_end(l_ap);
      return false;
    } break;
    case -1 : {
      va_end(l_ap);
      return false;
    } break;
    default : {
      l_map = (*l_map)[l_str]->GetMap();
    }
    };
  }
  va_end(l_ap);
  __current_map = l_map;
  return true;
}

int
Config::_isMap(Conf_Map_t* iMap, string iName) {
  if (iMap->find(iName) == iMap->end()) { // change to find
    return -1;
  } else {
    if ((*iMap)[iName]->IsMap() == true) {
      return 1;
    } else {
      return 0;
    }
  }
}

string
token(const string iString, const string iTokenSeparators, 
      int& iStrPos, bool iChew = true) {
  // iChew set to true eats token markers between tokens
  // so that iStrPos starts on next token
  int    l_end_pos;
  int    l_start_pos;
  string l_token;
  
  if (iStrPos >= iString.size()) {
    iStrPos = string::npos;
    return "";
  }
  l_start_pos = iString.find_first_not_of(iTokenSeparators, iStrPos);
  l_end_pos = iString.find_first_of(iTokenSeparators, l_start_pos);
  if (l_end_pos != string::npos) {
    l_token = iString.substr(l_start_pos, l_end_pos - l_start_pos);
    if (iChew) {
      iStrPos = iString.find_first_not_of(iTokenSeparators, l_end_pos);
      if (iStrPos == string::npos) {
	iStrPos = iString.size()+1;
      }
    } else {
      iStrPos = l_end_pos;
    }
    return l_token;
  } else {
    if (l_start_pos < iString.size()) {
      iStrPos = iString.size()+1;
      return iString.substr(l_start_pos, string::npos);
    } else {
      iStrPos = string::npos;
      return "";
    }
  }
}

void
Config::_read_level(ifstream &iFile, Conf_Map_t* iLevel, string iLevelName) {
  Conf_Map_t* l_level;
  Conf_Map_t* l_new_level;
  ConfigItem* l_new_item;
  string      l_line;
  string      l_token;
  string      l_value;
  string*     l_str_ptr;
  int         l_pos;

  while(!iFile.eof()) {
    __line++;
    __DEBUG << "\nchecking line " << __line << ": ";
    getline(iFile, l_line);
    l_pos = l_line.find_first_of("#");
    if (l_pos != string::npos) {
      l_line.erase(l_pos, string::npos);
    }
    l_pos = 0;
    l_token = token(l_line, " \t>", l_pos);
    if (l_pos != string::npos) {
      l_value = token(l_line, ">", l_pos);
      if (l_pos == string::npos) {
	// we have a boolean value, since it is in the config mark it on
	l_value = "on";
      }
      if (l_token[0] == '<') { // level change
	if (l_token[1] == '/') { // close level
	  // check for match?
	  if (l_token.substr(2, string::npos) != iLevelName) {
	    cerr << "Error line " << __line << ": level type doesn't match, opened " <<
	      iLevelName << " closed " << l_token.substr(2, string::npos) << "\n";
	  } else {
	    __DEBUG << "closing level " << l_token.substr(2, string::npos);
	  }
	  return;
	} else { // open a possibly new level
	  l_token = l_token.substr(1, string::npos);
	  if (iLevel->find(l_token) != iLevel->end()) {
	    if ((*iLevel)[l_token]->IsMap()) {
	      __DEBUG << "adding to level " << l_token << " => " << l_value;
	      l_new_level = new Conf_Map_t;
	      l_level = (*iLevel)[l_token]->GetMap();
	      l_new_item = new ConfigItem(l_new_level);
	      (*l_level)[l_value] = l_new_item;
	      _read_level(iFile, l_new_level, l_token);
	    } else {
	      cerr << "Error line " << __line << ": Token " <<
		l_token << " already exists as a key value pair\n";
	    }
	  } else { // entry doesn't exist at all
	    __DEBUG << "adding new level " << l_token << " => " << l_value;
	    l_new_level = new Conf_Map_t;
	    l_new_item = new ConfigItem(l_new_level);
	    (*iLevel)[l_token] = l_new_item;
	    l_level = l_new_level;
	    l_new_level = new Conf_Map_t;
	    l_new_item = new ConfigItem(l_new_level);
	    (*l_level)[l_value] = l_new_item;
	    _read_level(iFile, l_new_level, l_token);
	  }
	}
      } else { // regular key value pair
	__DEBUG << "adding " << l_token << " => " << l_value;
	l_str_ptr = new string(l_value);
	l_new_item = new ConfigItem((void *)l_str_ptr);
	(*iLevel)[l_token] = l_new_item;
      }
    } else {
      __DEBUG << "No token on line " << __line;
    }
  }
}

void
Config::LoadFile(const char* iFilename) {
  ifstream l_file(iFilename);

  if (l_file == NULL) {
    cerr << "Can't open file ";
    perror(iFilename);
    exit(errno);
  }
  __line = 0;
  __current_map = &__root;
  _read_level(l_file, __current_map, "__ROOT__");
}

Config::Config() {
  __current_map = &__root;
  __filename = 0; 
  __debug_ptr = &__debug_null;
}

Config::Config(const char* iFilename, bool iDebug = false) {
  if (iDebug) {
    __debug_ptr = &cout;
  } else {
    __debug_ptr = &__debug_null;
  }
  LoadFile(iFilename);
}

Config::~Config() {
  // destroy allocated map stuff
}

void
Config::_dump_level(Conf_Map_t* iLevel, string iTab) {
  Conf_Map_t::const_iterator l_iterate;

  for (l_iterate = iLevel->begin(); l_iterate != iLevel->end(); l_iterate++) {
    cout << iTab << l_iterate->first << "\t" << l_iterate->second << "\n";
    if (_isMap(iLevel, l_iterate->first)) {
      _dump_level((l_iterate->second)->GetMap(), iTab + "\t");
    }
  }
}

void
Config::DebugDump() {
  _dump_level(__current_map, "");
}
