/*
 * ipurge
 *
 * delete mail from cyrus imap mailbox or partition
 * based on date (or size?)
 *
 * $Id: ipurge.c,v 1.6 1998/09/29 16:20:11 eric Exp $
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <syslog.h>

/* cyrus includes */
#include "mailbox.h"

/* globals for getopt routines */
extern char *optarg;
extern int  optind;
extern int  opterr;
extern int  optopt;

/* globals for callback functions */
int days = -1;
int size = -1;
int exact = -1;

int purge_me(char *, int, int);
int purge_check(void *, char *);

int
main (int argc, char *argv[]) {
  char option;
  char buf[MAX_MAILBOX_PATH];

  if (geteuid() == 0) { /* don't run as root, changes permissions */
    usage(argv[0]);
  }

  while ((option = getopt(argc, argv, "hxd:b:k:m:")) != EOF) {
    switch (option) {
    case 'd': {
      if (optarg == 0) {
	usage(argv[0]);
      }
      days = atoi(optarg) * 86400 /* nominal # of seconds in a 'day' */;
    } break;
    case 'b': {
      if (optarg == 0) {
	usage(argv[0]);
      }
      size = atoi(optarg);
    } break;
    case 'k': {
      if (optarg == 0) {
	usage(argv[0]);
      }
      size = atoi(optarg) * 1024; /* make it bytes */
    } break;
    case 'm': {
      if (optarg == 0) {
	usage(argv[0]);
      }
      size = atoi(optarg) * 1048576; /* 1024 * 1024 */
    } break;
    case 'x' : {
      exact = 1;
    } break;
    case 'h':
    default: usage(argv[0]);
    }
  }
  if ((days == -1 ) && (size == -1)) {
    printf("One of these must be specified -d, -b -k, -m\n");
    usage(argv[0]);
  }
  config_init("IPurge");
  if (optind == argc) { /* do the whole partition */
    strcpy(buf, "*");
    mboxlist_findall(buf, 1, 0, 0, purge_me);
  } else {
    for (; optind < argc; optind++) {
      purge_me(argv[optind], 0, 0);
    }
  }
}

int
usage(char *name) {
  printf("usage: %s [-x] {-d days &| -b bytes|-k Kbytes|-m Mbytes} [mail_login]\n", name);
  printf("\tthere are no defaults and at least one of -d, -b, -k, -m must be specified\n");
  printf("\tif mail_login is not given %s works on the whole partition\n");
  printf("\t -x specifies an exact match for days or size\n");
  printf("\t***IMPORTANT*** run this as the cyrus owner\n");
  exit(0);
}

/* we don't check what comes in on matchlen and maycreate, should we? */
int
purge_me(char *name, int matchlen, int maycreate) {
  struct mailbox the_box;
  int            error;

  error = mailbox_open_header(name, 0, &the_box);
  if (error != 0) { /* did we find it? */
    syslog(LOG_ERR, "Couldn't find %s, check spelling (user.????)", name);
    return error;
  }
  if (the_box.header_fd != -1) {
    (void) mailbox_lock_header(&the_box);
  }
  the_box.header_lock_count = 1;

  error = chdir(the_box.path);
  if (error < 0) {
    syslog(LOG_ERR, "Couldn't change directory to %s : %m", the_box.path);
    return error;
  }
  error = mailbox_open_index(&the_box);
  if (error != 0) {
    mailbox_close(&the_box);
    syslog(LOG_ERR, "Couldn't open mailbox index for %s", name);
    return error;
  }
  (void) mailbox_lock_index(&the_box);
  the_box.index_lock_count = 1;

  mailbox_expunge(&the_box, 1, purge_check, 0);
  mailbox_close(&the_box);
  return 0;
}

/* thumbs up routine, checks date & size and returns yes or no for deletion */
/* 0 = no, 1 = yes */
int
purge_check(void *deciderock, char *buf) {
  struct index_record *the_record;
  unsigned long       my_time;

  the_record = (struct index_record *)buf;

  if (exact == 1) {
    if (days >= 0) {
      my_time = time(0);
      /*    printf("comparing %ld :: %ld\n", my_time, the_record->sentdate); */
      if (((my_time - the_record->sentdate)/86400) == (days/86400)) {
	return 1;
      }
    }
    if (size >= 0) {
      /* check size */
      if (the_record->size == size) {
	return 1;
      }
    }
    return 0;
  } else {
    if (days >= 0) {
      my_time = time(0);
      /*    printf("comparing %ld :: %ld\n", my_time, the_record->sentdate); */
      if ((my_time - the_record->sentdate) > days) {
	return 1;
      }
    }
    if (size >= 0) {
      /* check size */
      if (the_record->size > size) {
	return 1;
      }
    }
    return 0;
  }
}

/* fatal needed for imap library */
int
fatal(char *s, int code) {
  syslog(LOG_ERR, "Fatal error : %s", s);
  exit(code);
}
