/*
 * BerkeleyDB authentication module for ProFTPD
 *
 * $Id: mod_dbauth.c,v 1.7 1999/04/09 15:41:32 eric Exp $
 * Copyright 1999 Eric Estabrooks
 *
 * This module is free to redistribute and/or modify.  If you do modify it 
 * note the modifications before redistributing it.
 * 
 * This modules comes with no warranty either expressed or implied.  I am not
 * responsible for any damages that may result from the use of this module,
 * including but not limited to monetary loss, mental distress, or general
 * mayhem
 */

/* 
   Information for Configure, This should be modified to the appropriate
   location for your BerkeleyDB libraries and headers you will probably have
   to modify your Makes.rules.in (the CPPFLAGS line) to include the path
   to your BerkeleyDB headers ( -I/path/to/headers ).  The line below is 
   automatically pulled in by configure so you need to modify it to have
   the correct location of you library

   $Libraries: -L/usr/local/lib -ldb$

*/

/* 
   This module requires BerkeleyDB libraries to compile

   Configuration Options:

   DBAuth on               <- enables DBAuthentication
   DBAuthUserFile filename <- DBFile to check information against

*/


#include "conf.h"
#include "privs.h"

#include "db.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <crypt.h>

static char *str_strtok(char *s1, const char *s2) {
  static char* str_hold = 0;
  static char* cur_pos = 0;
  char*        new_pos;
  char*        old_pos;

  if (s2 == (char *)0) {
    return 0;
  }
  if (s1 != (char *)0) {
    str_hold = s1;
    cur_pos = s1;
  }
  old_pos = cur_pos;
  new_pos = strstr(cur_pos, s2);
  if (new_pos != 0) {
    *new_pos = 0;
    cur_pos = new_pos + strlen(s2);
    return old_pos;
  } else {
    if ((old_pos != (char *)0) && (strlen(old_pos) != 0)) {
      return old_pos;
    } else {
      return 0;
    }
  }
}

MODRET getpwnam_db(cmd_rec *cmd) {
  static struct passwd pass;
  static char          pw_name[128];
  static char          pw_passwd[64];
  static char          pw_gecos[1024];
  static char          pw_dir[1024];
  static char          pw_shell[256];
  struct passwd        *my_pass;
  DB_ENV               db_env;
  DB_INFO              db_info;
  DB                   *my_db;
  DBT                  query;
  DBT                  ans;
  char                 *buf;
  char                 *tmp; /* for robustness */
  char                 tmpans[4096];
  int                  error;
  int                  dbauth_on;
  char                 *dbfile = 0;
  int                  dbfallback = 0;
 
  dbauth_on = get_param_int(CURRENT_CONF, "DBAuth", FALSE);
  if (dbauth_on != TRUE) { /* are we even doing DB Authentication */
    return DECLINED(cmd);
  }
  dbfile = get_param_ptr(CURRENT_CONF, "DBAuthUserFile", FALSE);
  if (dbfile == NULL) {
    log_pri(LOG_ERR, "DBAuthUserFile not specified");
    return ERROR_MSG(cmd, R_421, "Service not available");
  }
  dbfallback = get_param_int(CURRENT_CONF, "DBAuthFail", FALSE);
  log_debug(DEBUG3, "dbfallback = %d\n", dbfallback);
  if (dbfallback == -1) {
    dbfallback = 0;
  }
	
  memset((void *)&db_env, 0, sizeof(db_env));
  memset((void *)&db_info, 0, sizeof(db_info));
  memset((void *)&query, 0, sizeof(query));
  memset((void *)&ans, 0, sizeof(ans));
  memset((void *)tmpans, 0, 4096);
  if (strlen(cmd->argv[0]) > 32) { /* someone is trying to hack */
    if (dbfallback) {
      return DECLINED(cmd);
    }
    add_response_err(R_DUP,"Illegal username");
    return ERROR(cmd);
  }
  PRIVS_ROOT
  error = db_open(dbfile, DB_HASH, DB_RDONLY, 0, &db_env, &db_info, &my_db);
  if (error != 0) {
    PRIVS_RELINQUISH
    log_pri(LOG_CRIT, "Couldn't open DBAuthUserFile [%s] : %s",
	    dbfile, strerror(errno));
    if (dbfallback) {
      return DECLINED(cmd);
    }
    return ERROR(cmd);
  }
  query.data = (void *)cmd->argv[0];
  query.size = strlen(cmd->argv[0]);
  ans.data = tmpans;
  ans.flags |= DB_DBT_USERMEM;
  ans.ulen = 4095;
  
  error = my_db->get(my_db, 0, &query, &ans, 0);
  if (error != 0) {
    errno = error;
    my_db->close(my_db,0);
    PRIVS_RELINQUISH
    if (dbfallback) {
      return DECLINED(cmd);
    }
    return ERROR_INT(cmd, AUTH_NOPWD);
  }
  tmpans[ans.size] = 0;
  tmpans[4095] = 0;
  my_db->close(my_db,0);
  PRIVS_RELINQUISH

  /* fill passwd structure */
  memset((void *)&pass, 0, sizeof(pass));
  memset((void *)pw_name, 0, 128);
  memset((void *)pw_passwd, 0, 64);
  memset((void *)pw_gecos, 0, 1024);
  memset((void *)pw_shell, 0, 256);
  memset((void *)pw_dir, 0, 1024);

  my_pass = &pass;
  my_pass->pw_name = pw_name;
  my_pass->pw_passwd = pw_passwd;
  my_pass->pw_gecos = pw_gecos;
  my_pass->pw_dir = pw_dir;
  my_pass->pw_shell = pw_shell;
  my_pass->pw_uid = -1;
  my_pass->pw_gid = -1;
  /* 
     not used
     my_pass->pw_age;
     my_pass->pw_comment;
   */
  tmp = (char *)str_strtok(tmpans, ":");
  if (tmp != (char *)0) {
    strcpy(pw_name, tmp);
    tmp = (char *)str_strtok(0,":");
    if (tmp != (char *)0) {
      strcpy(pw_passwd, tmp);
      tmp = (char *)str_strtok(0,":");
      if (tmp != (char *)0) {
        my_pass->pw_uid = atoi(tmp);
        tmp = (char *)str_strtok(0,":");
        if (tmp != (char *)0) {
          my_pass->pw_gid = atoi(tmp);
          tmp = (char *)str_strtok(0,":");
          if (tmp != (char *)0) {
            strcpy(pw_gecos, tmp);
            tmp = (char *)str_strtok(0,":");
            if (tmp != (char *)0) {
              strcpy(pw_dir, tmp);
              tmp = (char *)str_strtok(0,":");
	      if (tmp != (char *)0) {
		strcpy(pw_shell, tmp);
	      }
	    }
	  }
	}
      }
    }
  }
  log_debug(DEBUG3, "user found [%s] at [%s]", my_pass->pw_name, my_pass->pw_dir);
  return mod_create_data(cmd, my_pass);
}

MODRET pw_auth_db(cmd_rec *cmd) {
  modret_t *ret_ptr;
  struct passwd *my_pass;

  ret_ptr = getpwnam_db(cmd);
  if (MODRET_ISHANDLED(ret_ptr) && MODRET_HASDATA(ret_ptr)) {
    my_pass = (struct passwd *)ret_ptr->data;
  } else {
    return ret_ptr;
  }
  if (strcmp(my_pass->pw_passwd, crypt(cmd->argv[1], my_pass->pw_passwd)) == 0) {
    return HANDLED(cmd);
  } else {
    return ERROR_INT(cmd, AUTH_BADPWD);
  }
}
  
/*
  configuration routines
*/

MODRET add_dbauth(cmd_rec *cmd) {
  int        value;
  config_rec *config_ptr;

  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL);
  
  value = get_boolean(cmd, 1);
  if (value == -1) {
    CONF_ERROR(cmd, "requires a boolean value");
  }
  config_ptr = add_config_param("DBAuth", 1, (void *)value);
  config_ptr->flags |= CF_MERGEDOWN;

  return HANDLED(cmd);
}

MODRET add_dbauth_file(cmd_rec *cmd) {
  config_rec   *config_ptr;

  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL);
  
  log_debug(DEBUG3, "added dbauth_file [%s]", cmd->argv[1]);
  config_ptr = add_config_param_str("DBAuthUserFile", 1, (void *)cmd->argv[1]);
  config_ptr->flags |= CF_MERGEDOWN;

  return HANDLED(cmd);
}

MODRET add_dbauth_fail(cmd_rec *cmd) {
  int        value = 0;
  config_rec *config_ptr;

  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL);
  
  if (!strcasecmp(cmd->argv[1], "error")) {
    value = 0;
  } else {
    if (!strcasecmp(cmd->argv[1], "fallback")) {
      value = 1;
    } else {
      CONF_ERROR(cmd, "DBAuthFail can only be set to 'error' or 'fallback'");
    }
  }
  config_ptr = add_config_param("DBAuthFail", 1, (void *)value);
  config_ptr->flags |= CF_MERGEDOWN;

  return HANDLED(cmd);
}

/*
  module interface information
*/
static conftable dbauth_conftable[] = {
  { "DBAuth",          add_dbauth,      NULL }, /* turn BerkeleyDB Auth on and off */
  { "DBAuthUserFile",  add_dbauth_file, NULL }, /* hash file to use for the database */
  { "DBAuthFail",      add_dbauth_fail, NULL }, /* fallback if db failure
						   value can be "error" or "fallback" */
  { NULL }
};

static authtable dbauth_authtable[] = {
  { 0, "getpwnam", getpwnam_db },
  { 0, "auth",     pw_auth_db },
  { 0,        NULL }
};

module dbauth_module = {
  NULL,NULL,                    /* Always NULL */
  0x20,                         /* API Version 2.0 */
  "dbauth",
  dbauth_conftable,              /* dbauth configuration handler table */
  NULL,                          /* dbauth command handler table */
  dbauth_authtable,              /* No authentication handler table */
  NULL,                          /* Initialization function */
  NULL,                          /* Post-fork "child mode" init */
};
