/*
 * OPIE authentication module for ProFTPD
 *
 * $Id: mod_opie.c,v 1.1 1999/02/13 17:40:03 eric Exp $
 * Copyright 1999 Eric Estabrooks
 *
 * This module is free to redistribute and/or modify.  If you do modify it 
 * note the modifications before redistributing it.
 * 
 * This modules comes with no warranty either expressed or implied.  I am not
 * responsible for any damages that may result from the use of this module,
 * including but not limited to monetary loss, mental distress, or general
 * mayhem
 */

/* 
   Information for Configure, This should be modified to the appropriate
   location for your OPIE library and headers you will probably have
   to modify your Makes.rules.in (the CPPFLAGS line) to include the path
   to your OPIE headers ( -I/path/to/headers ).  The line below is 
   automatically pulled in by configure so you need to modify it to have
   the correct location of you library

   $Libraries: -L/usr/local/opie/lib -lopie$

*/

/* 
   This module requires OPIE libraries to compile
   http://www.inner.net is where I got my copy of OPIE

   This module allows users who are not in the opie database to authenticate
   via the usual methods.

   Configuration Options:

   OPIEAuth on               <- enables OPIE Authentication
   
*/

#include "conf.h"
#include "privs.h"

#include "opie_cfg.h"
#include "opie.h"

#include <sys/types.h>
#include <sys/stat.h>

MODRET post_cmd_user(cmd_rec *cmd) {
  struct opie my_opie;
  int         result;
  int         opie_auth_on;
  char        challenge[OPIE_CHALLENGE_MAX+1];
  char        user[128];

  opie_auth_on = get_param_int(CURRENT_CONF, "OPIEAuth", FALSE);
  if (opie_auth_on != TRUE) { /* are we doing opie */
    return DECLINED(cmd);
  }
  memset((void *)user, 0, 128);
  strncpy(user, cmd->argv[1], 127);
  PRIVS_ROOT
  result = opiechallenge(&my_opie, user, challenge);
  PRIVS_RELINQUISH
  log_debug(DEBUG3, "opie challenge: %s => %s [%d]", user, challenge, result);
  if (result == 0) {
      add_response(R_331, challenge);
  }
  return DECLINED(cmd);
}

MODRET opie_auth(cmd_rec *cmd) {
  struct opie my_opie;
  int         i;
  int         result;
  int         opie_auth_on;
  char        user[128];
  char        response[OPIE_RESPONSE_MAX+1];
  char        challenge[OPIE_CHALLENGE_MAX+1];

  opie_auth_on = get_param_int(CURRENT_CONF, "OPIEAuth", FALSE);
  if (opie_auth_on != TRUE) {
    return DECLINED(cmd);
  }
  memset((void *)user , 0, 128);
  strncpy(user, cmd->argv[0], 127);
  PRIVS_ROOT
  result = opiechallenge(&my_opie, user, challenge);
  PRIVS_RELINQUISH
  if (result == 1) { /* normal users fall through */
    return DECLINED(cmd);
  }
  memset((void *)response, 0, OPIE_RESPONSE_MAX+1);
  /* copy otp response and convert to upper case */
  for (i=0; (i < OPIE_RESPONSE_MAX) || (*(cmd->argv[1]+i) == 0); i++) {
    response[i] = toupper(*(cmd->argv[1]+i));
  }
  PRIVS_ROOT
  result = opieverify(&my_opie, response);
  PRIVS_RELINQUISH
  log_debug(DEBUG3, "[%s] => %s [%d]", challenge, response, result);
  if (result == 0) {
    return HANDLED(cmd);
  } else {
    return ERROR_INT(cmd, AUTH_BADPWD);
  }
}
  
/*
  configuration routines
*/

MODRET add_opie(cmd_rec *cmd) {
  int        value;
  config_rec *config_ptr;

  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL);
  
  value = get_boolean(cmd, 1);
  if (value == -1) {
    CONF_ERROR(cmd, "requires a boolean value");
  }
  config_ptr = add_config_param("OPIEAuth", 1, (void *)value);
  config_ptr->flags |= CF_MERGEDOWN;

  return HANDLED(cmd);
}

/*
  module interface information
*/
static conftable opie_conftable[] = {
  { "OPIEAuth", add_opie, NULL }, /* turn OPIE Auth on and off */
  { NULL }
};

static cmdtable opie_cmdtable[] = {
  { POST_CMD, C_USER, G_NONE, post_cmd_user, FALSE, FALSE },
  { 0,         NULL }
};

static authtable opie_authtable[] = {
  { 0, "auth", opie_auth },     /* verifies response to challenge */
  { 0, NULL }
};

module opie_module = {
  NULL,NULL,                    /* Always NULL */
  0x20,                         /* API Version 2.0 */
  "opie_auth",
  opie_conftable,               /* opie configuration handler table */
  opie_cmdtable,                /* opie command handler table */
  opie_authtable,               /* authentication handler table */
  NULL,                         /* Initialization function */
  NULL,                         /* Post-fork "child mode" init */
};
