/*
 * msaslpasswd - interim saslpasswd that does it's thing to both 
 *               v1 and v2 sasldb files
 *               it's just both saslpasswds wrapped in a dynamic load
 * 
 */

#include <dlfcn.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>


#define Using_BDB 1
#ifdef Using_BDB
#include <db.h>
void stupid_pet_tricks(void);
#endif

/* various sasl defines */
#define SASL_FAIL -1
#define SASL_OK    0

extern char *optarg;
extern int optind, opterr, optopt;

/* my sasl library had this as an undefined
   symbol under solaris, couldn't find a 
   library that contained it, oh well
*/
void __eprintf(void) {
}

int
main (int argc, char* argv[]) {
  char* progname = NULL;
  char  myhostname[1025];

  void* v1_handle;
  void* v2_handle;
  int (*v1_sasl_pass)(char* host, char* user, char* pass,
                      unsigned int len, char* domain, char* appname,
                      int flag_create, int flag_disable);
  
  /* v2 evil, these are interal functions getting called at the api layer, sadness */
  int (*v2_sasl_pass)(char* host, char* user, char* pass,
                      unsigned int len, char* domain, char* appname,
                      int flag_create, int flag_disable);
  void (*v1_read_pass)(char* prompt, int flag_pipe, char** passwd, unsigned int* len);
  
  int err;

  /* saslpasswd_v1 parms */
  int flag_pipe = 0, flag_create = 0, flag_disable = 0, flag_error = 0;
  int c;
  char *userid, *password, *verify;
  unsigned passlen, verifylen;
  int result;
  char *user_domain = NULL;
  char *appname = "saslpasswd";

  /* saslpasswd_v2 parms and some code */
  memset(myhostname, 0, sizeof(myhostname));
  result = gethostname(myhostname, sizeof(myhostname)-1);
  if (result == -1) {
    perror("gethostname failed");
    exit(SASL_FAIL);
  }

#ifdef Using_BDB
   stupid_pet_tricks(); /* we have a hardcoded db name here */
#endif


  /* option parsing code from saslpasswd_v1 */
  if (! argv[0])
    progname = "saslpasswd";
  else {
    progname = strrchr(argv[0], '/');
    if (progname)
      progname++;
    else
      progname = argv[0];
  }

  while ((c = getopt(argc, argv, "pcdu:a:h?")) != EOF)
    switch (c) {
    case 'p':
      flag_pipe = 1;
      break;
    case 'c':
      if (flag_disable)
        flag_error = 1;
      else
        flag_create = 1;
      break;
    case 'd':
      if (flag_create)
        flag_error = 1;
      else
        flag_disable = 1;
      break;
    case 'u':
      user_domain = optarg;
      break;
    case 'a':
      appname = optarg;
      if (strchr(optarg, '/') != NULL) {
        (void)fprintf(stderr, "filename must not contain /\n");
        exit(-SASL_FAIL);
      }
      break;
    default:
      flag_error = 1;
      break;
    }

  if (optind != argc - 1)
    flag_error = 1;

  if (flag_error) {
    (void)fprintf(stderr,
                  "%s: usage: %s [-a appname] [-p] [-c] [-d] [-u DOM] userid\n"
                  "\t-p\tpipe mode -- no prompt, password read on stdin\n"
                  "\t-c\tcreate -- ask mechs to create the account\n"
                  "\t-d\tdisable -- ask mechs to disable the account\n"
                  "\t-a appname\tuse appname as application name\n"
                  "\t-u DOM\tuse DOM for user domain\n",
                  progname, progname);
    exit(-SASL_FAIL);
  }

  userid = argv[optind];

  v1_handle = dlopen("libmpass_sasl1.so", RTLD_NOW);
  if (v1_handle == NULL) {
    fprintf(stderr, "%s\n", dlerror());
    exit(-3);
  }
  v2_handle = dlopen("libmpass_sasl2.so", RTLD_NOW);
  if (v2_handle == NULL) {
    fprintf(stderr, "%s\n", dlerror());
    dlclose(v1_handle);
    exit(-4);
  }

  /* load all the symbols we are going to use */
  v1_sasl_pass = dlsym(v1_handle, "saslv1_evil");
  if (v1_sasl_pass == NULL) {
    fprintf(stderr, "saslv1_evil was undefined in libmpass_sasl1.so: %s\n",
            dlerror());
    goto leaving_error;
  }
  v1_read_pass = dlsym(v1_handle, "read_password");
  if (v1_read_pass == NULL) {
    fprintf(stderr, "read_password was undefined in libmpass_sasl1.so: %s\n",
            dlerror());
    goto leaving_error;
  }
  v2_sasl_pass = dlsym(v2_handle, "saslv2_evil");
  if (v2_sasl_pass == NULL) {
    fprintf(stderr, "saslv2_evil was undefined in libmpass_sasl2.so: %s\n",
            dlerror());
    goto leaving_error;
  }
  
  /* more saslpasswd v1 code */
  if (!flag_disable) {
    (*v1_read_pass)("Password: ", flag_pipe, &password, &passlen);
    
    if (! flag_pipe) {
      (*v1_read_pass)("Again (for verification): ", flag_pipe, &verify,
                    &verifylen);
      if (passlen != verifylen
          || memcmp(password, verify, verifylen)) {
        fprintf(stderr, "%s: passwords don't match; aborting\n", 
                progname);
        exit(-SASL_FAIL);
      }
    }
  }

  /* should auto it here, but stuck under solaris argh */
  putenv("SASL_PATH=/usr/lib/sasl");  
  err = (*v1_sasl_pass)(myhostname, userid, password, passlen,
                        user_domain, appname, flag_create, 
                        flag_disable);
  if (err == SASL_OK) {
    printf("successfully updated sasldb v1\n");
  } else {
    printf("failed to update sasldb v1 [%d]\n", err);
    goto leaving_error;
  }

  /* some evil that v2 does with internal functions rather than a nice
     clean function call so a function was added to saslpasswd2.c 
     saslv2_evil to provide this.
  */
  putenv("SASL_PATH=/usr/lib/sasl2");
  err = (*v2_sasl_pass)(myhostname, userid, password, passlen,
                        user_domain, appname, flag_create,
                        flag_disable);
  if (err == SASL_OK) {
    printf("successfully updated sasldb v2\n");
  } else {
    printf("failed to update sasldb v2, you now have an anomaly between sasldbs\n");
    goto leaving_error;
  }

  dlclose(v1_handle);
  dlclose(v2_handle);
  return SASL_OK;
  
 leaving_error:
   dlclose(v1_handle);
  dlclose(v2_handle);
  return SASL_FAIL;
 
}

#ifdef Using_BDB
void
stupid_pet_tricks(void) {
   DB*   mbdb;
   int   ret;
   char* path = "/etc/sasldb";

#if DB_VERSION_MAJOR < 3
   ret = db_open(path, DB_HASH, DB_CREATE, 0660, NULL, NULL, &mbdb);
#else
    ret = db_create(&mbdb, NULL, 0);
    if (ret == 0 && mbdb != NULL)
    {
            ret = (mbdb)->open(mbdb, path, NULL, DB_HASH, DB_CREATE, 0660);
            if (ret != 0)
            {
                    (void) (mbdb)->close(mbdb, 0);
                    mbdb = NULL;
            }
    }
#endif
 mbdb->close(mbdb,0); 
}
#endif
